/*
 Copyright (C) 2006-2018 M.A.L. Marques
 Copyright (C) 2019 X. Andrade

 This Source Code Form is subject to the terms of the Mozilla Public
 License, v. 2.0. If a copy of the MPL was not distributed with this
 file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/

/**
 * @file work_gga.c
 * @brief This file is to be included in GGA functionals.
 */

#ifdef XC_DEBUG
#define __USE_GNU
#include <fenv.h>
#endif
#include <omp.h>

//#define USE_INTERNAL_COUNTERS

/**
 * @param[in,out] func_type: pointer to functional structure
 */
static void
work_gga_genxc(const XC(func_type) *p, size_t np,
#ifdef USE_INTERNAL_COUNTERS
         const double *irho, const double *isigma, double *izk,
         double *ivrho, double *ivsigma, double *iv2rho2, double *iv2rhosigma, double *iv2sigma2,
         double *iv3rho3, double *iv3rho2sigma, double *iv3rhosigma2, double *iv3sigma3,
         double *iv4rho4, double *iv4rho3sigma, double *iv4rho2sigma2, double *iv4rhosigma3, double *iv4sigma4
#else
         const double *rho, const double *sigma, double *zk,
         double *vrho, double *vsigma, double *v2rho2, double *v2rhosigma, double *v2sigma2,
         double *v3rho3, double *v3rho2sigma, double *v3rhosigma2, double *v3sigma3,
         double *v4rho4, double *v4rho3sigma, double *v4rho2sigma2, double *v4rhosigma3, double *v4sigma4
#endif
) {

  int order = -1;
#ifdef USE_INTERNAL_COUNTERS
  if(izk     != NULL) order = 0;
  if(ivrho   != NULL) order = 1;
  if(iv2rho2 != NULL) order = 2;
  if(iv3rho3 != NULL) order = 3;
  if(iv4rho4 != NULL) order = 4;
#else
  if(zk     != NULL) order = 0;
  if(vrho   != NULL) order = 1;
  if(v2rho2 != NULL) order = 2;
  if(v3rho3 != NULL) order = 3;
  if(v4rho4 != NULL) order = 4;
#endif
  if(order < 0) return;

  size_t ip;
  double dens;
  double my_rho[2] = {0.0, 0.0};
  double my_sigma[3] = {0.0, 0.0, 0.0};

  if(p->nspin == XC_POLARIZED) {
    #pragma omp parallel for private(ip, my_rho, my_sigma, dens)
    for(ip = 0; ip < np; ip++){

#ifndef USE_INTERNAL_COUNTERS
      const xc_dimensions *dim = &(p->dim);
      const int pos = ip;
      const double *irho = &rho[pos * dim->rho];
      const double *isigma = &sigma[pos * dim->sigma];
      double *izk = &zk[pos * dim->zk];
      double *ivrho = &vrho[pos * dim->vrho];
      double *ivsigma = &vsigma[pos * dim->vsigma];
      double *iv2rho2 = &v2rho2[pos * dim->v2rho2];
      double *iv2rhosigma = &v2rhosigma[pos * dim->v2rhosigma];
      double *iv2sigma2 = &v2sigma2[pos * dim->v2sigma2];
      double *iv3rho3 = NULL; 
      double *iv3rho2sigma = NULL; 
      double *iv3rhosigma2 = NULL; 
      double *iv3sigma3 = NULL; 
      if(v3rho3 != NULL) {
        iv3rho3 = &v3rho3[pos * dim->v3rho3];
        iv3rho2sigma = &v3rho2sigma[pos * dim->v3rho2sigma];
        iv3rhosigma2 = &v3rhosigma2[pos * dim->v3rhosigma2];
        iv3sigma3 = &v3sigma3[pos * dim->v3sigma3];
      }
      double *iv4rho4 = NULL;
      double *iv4rho3sigma = NULL;
      double *iv4rho2sigma2 = NULL;
      double *iv4rhosigma3 = NULL;
      double *iv4sigma4 = NULL;
      if(v4rho4 != NULL) {
        iv4rho4 = &v4rho4[pos * dim->v4rho4];
        iv4rho3sigma = &v4rho3sigma[pos * dim->v4rho3sigma];
        iv4rho2sigma2 = &v4rho2sigma2[pos * dim->v4rho2sigma2];
        iv4rhosigma3 = &v4rhosigma3[pos * dim->v4rhosigma3];
        iv4sigma4 = &v4sigma4[pos * dim->v4sigma4];
      }
#endif // USE_INTERNAL_COUNTERS

      dens = irho[0] + irho[1];
      if(dens >= p->dens_threshold) {
        my_rho[0] = m_max(p->dens_threshold, irho[0]);
        my_sigma[0] = m_max(p->sigma_threshold * p->sigma_threshold, isigma[0]);
        double s_ave;
        my_rho[1] = m_max(p->dens_threshold, irho[1]);
        my_sigma[2] = m_max(p->sigma_threshold * p->sigma_threshold, isigma[2]);
        my_sigma[1] = isigma[1];
        s_ave = 0.5*(my_sigma[0] + my_sigma[2]);
        my_sigma[1] = (my_sigma[1] >= -s_ave ? my_sigma[1] : -s_ave);
        my_sigma[1] = (my_sigma[1] <= +s_ave ? my_sigma[1] : +s_ave);
        func_pol(p, order, my_rho, my_sigma, izk, ivrho, ivsigma, iv2rho2, iv2rhosigma, iv2sigma2, iv3rho3, iv3rho2sigma, iv3rhosigma2, iv3sigma3, iv4rho4, iv4rho3sigma, iv4rho2sigma2, iv4rhosigma3, iv4sigma4);
      }
#ifdef USE_INTERNAL_COUNTERS
      internal_counters_gga_next(&(p->dim), 0, &irho, &isigma, &izk, &ivrho, &ivsigma, &iv2rho2, &iv2rhosigma, &iv2sigma2, &iv3rho3, &iv3rho2sigma, &iv3rhosigma2, &iv3sigma3, &iv4rho4, &iv4rho3sigma, &iv4rho2sigma2, &iv4rhosigma3, &iv4sigma4);
#endif
    }
  } else {
    #pragma omp parallel for
    for(ip = 0; ip < np; ip++){
#ifndef USE_INTERNAL_COUNTERS
      const xc_dimensions *dim = &(p->dim);
      const int pos = ip;
      const double *irho = &rho[pos * dim->rho];
      const double *isigma = &sigma[pos * dim->sigma];
      double *izk = &zk[pos * dim->zk];
      double *ivrho = &vrho[pos * dim->vrho];
      double *ivsigma = &vsigma[pos * dim->vsigma];
      double *iv2rho2 = &v2rho2[pos * dim->v2rho2];
      double *iv2rhosigma = &v2rhosigma[pos * dim->v2rhosigma];
      double *iv2sigma2 = &v2sigma2[pos * dim->v2sigma2];
      double *iv3rho3 = NULL; 
      double *iv3rho2sigma = NULL; 
      double *iv3rhosigma2 = NULL; 
      double *iv3sigma3 = NULL; 
      if(v3rho3 != NULL) {
        iv3rho3 = &v3rho3[pos * dim->v3rho3];
        iv3rho2sigma = &v3rho2sigma[pos * dim->v3rho2sigma];
        iv3rhosigma2 = &v3rhosigma2[pos * dim->v3rhosigma2];
        iv3sigma3 = &v3sigma3[pos * dim->v3sigma3];
      }
      double *iv4rho4 = NULL;
      double *iv4rho3sigma = NULL;
      double *iv4rho2sigma2 = NULL;
      double *iv4rhosigma3 = NULL;
      double *iv4sigma4 = NULL;
      if(v4rho4 != NULL) {
        iv4rho4 = &v4rho4[pos * dim->v4rho4];
        iv4rho3sigma = &v4rho3sigma[pos * dim->v4rho3sigma];
        iv4rho2sigma2 = &v4rho2sigma2[pos * dim->v4rho2sigma2];
        iv4rhosigma3 = &v4rhosigma3[pos * dim->v4rhosigma3];
        iv4sigma4 = &v4sigma4[pos * dim->v4sigma4];
      }
#endif // USE_INTERNAL_COUNTERS

      dens = irho[0];
      if(dens >= p->dens_threshold) {
        my_rho[0] = m_max(p->dens_threshold, irho[0]);
        my_sigma[0] = m_max(p->sigma_threshold * p->sigma_threshold, isigma[0]);
        func_unpol(p, order, my_rho, my_sigma, izk, ivrho, ivsigma, iv2rho2, iv2rhosigma, iv2sigma2, iv3rho3, iv3rho2sigma, iv3rhosigma2, iv3sigma3, iv4rho4, iv4rho3sigma, iv4rho2sigma2, iv4rhosigma3, iv4sigma4);
      }

#ifdef USE_INTERNAL_COUNTERS
      internal_counters_gga_next(&(p->dim), 0, &irho, &isigma, &izk, &ivrho, &ivsigma, &iv2rho2, &iv2rhosigma, &iv2sigma2, &iv3rho3, &iv3rho2sigma, &iv3rhosigma2, &iv3sigma3, &iv4rho4, &iv4rho3sigma, &iv4rho2sigma2, &iv4rhosigma3, &iv4sigma4);
#endif
    }
  }
}
